@@ -57,6 +57,13 @@
 #include "clutter-gst-player.h"
 #include "clutter-gst-private.h"
 
+#if defined (CLUTTER_WINDOWING_X11) && defined (HAVE_HW_DECODER_SUPPORT)
+#define GST_USE_UNSTABLE_API 1
+#include <gst/video/videocontext.h>
+#include <clutter/x11/clutter-x11.h>
+#endif
+
+
 typedef ClutterGstPlayerIface       ClutterGstPlayerInterface;
 
 G_DEFINE_INTERFACE_WITH_CODE (ClutterGstPlayer, clutter_gst_player, G_TYPE_OBJECT,
@@ -408,13 +415,18 @@ set_subtitle_uri (ClutterGstPlayer *player,
                   const gchar      *uri)
 {
   ClutterGstPlayerPrivate *priv = PLAYER_GET_PRIVATE (player);
+  GstPlayFlags flags;
 
   if (!priv->pipeline)
     return;
 
   CLUTTER_GST_NOTE (MEDIA, "setting subtitle URI: %s", uri);
 
+  g_object_get (priv->pipeline, "flags", &flags, NULL);
+
   g_object_set (priv->pipeline, "suburi", uri, NULL);
+
+  g_object_set (priv->pipeline, "flags", flags, NULL);
 }
 
 static void
@@ -1188,6 +1200,8 @@ on_volume_changed_main_context (gpointer data)
 
   g_object_notify (G_OBJECT (player), "audio-volume");
 
+  g_object_unref (player);
+
   return FALSE;
 }
 
@@ -1200,7 +1214,7 @@ on_volume_changed (GstElement       *pipeline,
 		   GParamSpec       *pspec,
 		   ClutterGstPlayer *player)
 {
-  g_idle_add (on_volume_changed_main_context, player);
+  g_idle_add (on_volume_changed_main_context, g_object_ref (player));
 }
 
 static GList *
@@ -1240,6 +1254,8 @@ on_audio_changed_main_context (gpointer data)
 
   g_object_notify (G_OBJECT (player), "audio-streams");
 
+  g_object_unref (player);
+
   return FALSE;
 }
 
@@ -1248,7 +1264,7 @@ static void
 on_audio_changed (GstElement       *pipeline,
                   ClutterGstPlayer *player)
 {
-  g_idle_add (on_audio_changed_main_context, player);
+  g_idle_add (on_audio_changed_main_context, g_object_ref (player));
 }
 
 static void
@@ -1263,7 +1279,7 @@ on_audio_tags_changed (GstElement       *pipeline,
   if (current_stream != stream)
     return;
 
-  g_idle_add (on_audio_changed_main_context, player);
+  g_idle_add (on_audio_changed_main_context, g_object_ref (player));
 }
 
 static gboolean
@@ -1274,6 +1290,8 @@ on_current_audio_changed_main_context (gpointer data)
   CLUTTER_GST_NOTE (AUDIO_STREAM, "audio stream changed");
   g_object_notify (G_OBJECT (player), "audio-stream");
 
+  g_object_unref (player);
+
   return FALSE;
 }
 
@@ -1282,7 +1300,7 @@ on_current_audio_changed (GstElement       *pipeline,
                           GParamSpec       *pspec,
                           ClutterGstPlayer *player)
 {
-  g_idle_add (on_current_audio_changed_main_context, player);
+  g_idle_add (on_current_audio_changed_main_context, g_object_ref (player));
 }
 
 static gboolean
@@ -1298,6 +1316,8 @@ on_text_changed_main_context (gpointer data)
 
   g_object_notify (G_OBJECT (player), "subtitle-tracks");
 
+  g_object_unref (player);
+
   return FALSE;
 }
 
@@ -1306,7 +1326,7 @@ static void
 on_text_changed (GstElement       *pipeline,
                   ClutterGstPlayer *player)
 {
-  g_idle_add (on_text_changed_main_context, player);
+  g_idle_add (on_text_changed_main_context, g_object_ref (player));
 }
 
 static void
@@ -1314,7 +1334,7 @@ on_text_tags_changed (GstElement       *pipeline,
                        gint              stream,
                        ClutterGstPlayer *player)
 {
-  g_idle_add (on_text_changed_main_context, player);
+  g_idle_add (on_text_changed_main_context, g_object_ref (player));
 }
 
 static gboolean
@@ -1325,6 +1345,8 @@ on_current_text_changed_main_context (gpointer data)
   CLUTTER_GST_NOTE (AUDIO_STREAM, "text stream changed");
   g_object_notify (G_OBJECT (player), "subtitle-track");
 
+  g_object_unref (player);
+
   return FALSE;
 }
 
@@ -1333,7 +1355,7 @@ on_current_text_changed (GstElement       *pipeline,
                           GParamSpec       *pspec,
                           ClutterGstPlayer *player)
 {
-  g_idle_add (on_current_text_changed_main_context, player);
+  g_idle_add (on_current_text_changed_main_context, g_object_ref (player));
 }
 
 /* GObject's magic/madness */
@@ -1835,6 +1857,7 @@ clutter_gst_player_set_subtitle_track_impl (ClutterGstPlayer *player,
                                             gint              index_)
 {
   ClutterGstPlayerPrivate *priv;
+  GstPlayFlags flags;
 
   priv = PLAYER_GET_PRIVATE (player);
 
@@ -1843,9 +1866,19 @@ clutter_gst_player_set_subtitle_track_impl (ClutterGstPlayer *player,
 
   CLUTTER_GST_NOTE (SUBTITLES, "set subtitle track to #%d", index_);
 
-  g_object_set (G_OBJECT (priv->pipeline),
-                "current-text", index_,
-                NULL);
+  g_object_get (priv->pipeline, "flags", &flags, NULL);
+  flags &= ~GST_PLAY_FLAG_TEXT;
+  g_object_set (priv->pipeline, "flags", flags, NULL);
+
+  if (index_ >= 0)
+    {
+      g_object_set (G_OBJECT (priv->pipeline),
+                    "current-text", index_,
+                    NULL);
+
+      flags |= GST_PLAY_FLAG_TEXT;
+      g_object_set (priv->pipeline, "flags", flags, NULL);
+    }
 }
 
 static gboolean
@@ -1860,6 +1893,38 @@ clutter_gst_player_get_idle_impl (ClutterGstPlayer *player)
 
 /**/
 
+#if defined (CLUTTER_WINDOWING_X11) && defined (HAVE_HW_DECODER_SUPPORT)
+static GstBusSyncReply
+on_sync_message (GstBus * bus, GstMessage * message, gpointer user_data)
+{
+  Display *display = user_data;
+  GstVideoContext *context;
+  const gchar **types;
+
+  if (gst_video_context_message_parse_prepare (message, &types, &context)) {
+    gint i;
+
+    for (i = 0; types[i]; i++) {
+
+      if (!strcmp(types[i], "x11-display")) {
+        gst_video_context_set_context_pointer (context, "x11-display", display);
+      }
+      else if (!strcmp(types[i], "x11-display-name")) {
+        gst_video_context_set_context_string (context, "x11-display-name",
+            DisplayString (display));
+      } else {
+        continue;
+      }
+
+      gst_message_unref (message);
+      return GST_BUS_DROP;
+    }
+  }
+
+  return GST_BUS_PASS;
+}
+#endif
+
 /**
  * clutter_gst_player_init:
  * @player: a #ClutterGstPlayer
@@ -1976,6 +2041,11 @@ clutter_gst_player_init (ClutterGstPlayer *player)
                     G_CALLBACK (on_current_text_changed),
                     player);
 
+#if defined(CLUTTER_WINDOWING_X11) && defined (HAVE_HW_DECODER_SUPPORT)
+  gst_bus_set_sync_handler (priv->bus, on_sync_message,
+      clutter_x11_get_default_display ());
+#endif
+
   gst_object_unref (GST_OBJECT (priv->bus));
 
   return TRUE;
